/**
  **************************************************************************
  * @file     at32_video_ev_lcd.c
  * @version  v2.0.2
  * @date     2022-09-15
  * @brief    at32_video_ev_lcd program
  **************************************************************************
  *                       Copyright notice & Disclaimer
  *
  * The software Board Support Package (BSP) that is made available to
  * download from Artery official website is the copyrighted work of Artery.
  * Artery authorizes customers to use, copy, and distribute the BSP
  * software and its related documentation for the purpose of design and
  * development in conjunction with Artery microcontrollers. Use of the
  * software is governed by this copyright notice and the following disclaimer.
  *
  * THIS SOFTWARE IS PROVIDED ON "AS IS" BASIS WITHOUT WARRANTIES,
  * GUARANTEES OR REPRESENTATIONS OF ANY KIND. ARTERY EXPRESSLY DISCLAIMS,
  * TO THE FULLEST EXTENT PERMITTED BY LAW, ALL EXPRESS, IMPLIED OR
  * STATUTORY OR OTHER WARRANTIES, GUARANTEES OR REPRESENTATIONS,
  * INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
  *
  **************************************************************************
  */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "at32_video_ev_lcd.h"
#include "at32_video_ev_font.h"
#include "at32_video_ev_spi.h"

/******************************************************************************/
/* point color and background color */
u16 POINT_COLOR = 0x0000;
u16 BACK_COLOR = 0xFFFF;

_lcd_dev lcddev;

/**
  * @brief  lcd reset
  * @param  none
  * @retval none
  */
void lcd_reset(void)
{
  LCD_RST_SET;
  delay_ms(30);
  LCD_RST_CLR;
  delay_ms(50);
  LCD_RST_SET;
  delay_ms(30);
}

#if(LCD_SPI1_EN)

/**
  * @brief  configures lcd control lines (fsmc pins) in alternate function push-pull mode
  * @param  none
  * @retval none
  */
void lcd_ctrllinesconfig(void)
{
  gpio_init_type	GPIO_InitStructure;

  crm_periph_clock_enable(CRM_GPIOA_PERIPH_CLOCK, TRUE);
  crm_periph_clock_enable(CRM_GPIOB_PERIPH_CLOCK, TRUE);
  crm_periph_clock_enable(CRM_GPIOC_PERIPH_CLOCK, TRUE);
  crm_periph_clock_enable(CRM_IOMUX_PERIPH_CLOCK, TRUE);

  gpio_pin_remap_config(SWJTAG_GMUX_010, TRUE);

  gpio_default_para_init(&GPIO_InitStructure);
  GPIO_InitStructure.gpio_pins = GPIO_PINS_15 | GPIO_PINS_9;
  GPIO_InitStructure.gpio_mode = GPIO_MODE_OUTPUT;
  gpio_init(GPIOA, &GPIO_InitStructure);

  GPIO_InitStructure.gpio_pins = GPIO_PINS_7; 		   //LCD_RS
  gpio_init(GPIOC, &GPIO_InitStructure);

  GPIO_InitStructure.gpio_pins = GPIO_PINS_0; 		   //LCD_CS
  gpio_init(GPIOB, &GPIO_InitStructure);

  gpio_bits_set(GPIOA, GPIO_PINS_9);
  gpio_bits_set(GPIOA, GPIO_PINS_15);
  gpio_bits_reset(GPIOB, GPIO_PINS_0);
}

/**
  * @brief  spi write display data
  * @param  dat: write data
  * @retval none
  */
void SPI4W_WriteDisplayData1(u16 dat)
{
  lcd_spi1_senddata(dat >> 8);
  lcd_spi1_senddata(dat & 0xFF);
}

#else//(LCD_SPI1_EN)

/**
  * @brief  configures lcd control lines (fsmc pins) in alternate function push-pull mode
  * @param  none
  * @retval none
  */
void lcd_ctrllinesconfig(void)
{

  gpio_init_type	GPIO_InitStructure;

  RCC_APB2PeriphClockCmd(RCC_APB2PERIPH_GPIOA | RCC_APB2PERIPH_GPIOB | RCC_APB2PERIPH_GPIOC, ENABLE);

  gpio_default_para_init(&GPIO_InitStructure);
  GPIO_InitStructure.gpio_pins = GPIO_PINS_7 | GPIO_PINS_5 | GPIO_PINS_9;
  GPIO_InitStructure.gpio_mode = GPIO_MODE_OUTPUT;
  gpio_init(GPIOA, &GPIO_InitStructure);

  GPIO_InitStructure.gpio_pins = GPIO_PINS_7; 		   //LCD_RS
  gpio_init(GPIOC, &GPIO_InitStructure);

  GPIO_InitStructure.gpio_pins = GPIO_PINS_6; 		   //LCD_CS
  gpio_init(GPIOB, &GPIO_InitStructure);

  gpio_bits_set(GPIOA, GPIO_PINS_9);
  gpio_bits_set(GPIOB, GPIO_PINS_6);
}

/**
  * @brief  wirte register index
  * @param  index: wirte register index
  * @retval none
  */
void spi4w_writeregindex(u16 Index)
{
  LCD_CS_CLR;
  LCD_RS_CLR;

  LCD_SCL_CLR;
  LCD_SDA_CLR;

  SPI_SDA = (Index >> 7);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 6);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 5);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 4);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 3);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 2);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 1);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (Index >> 0);
  LCD_SCL_CLR;
  LCD_SCL_SET;

  LCD_CS_SET;

}

/**
  * @brief  spi write data
  * @param  dat: write data
  * @retval none
  */
void spi4w_writedata(u16 dat)
{
  LCD_CS_CLR;
  LCD_RS_SET;

  LCD_SCL_CLR;
  LCD_SDA_SET;
  SPI_SDA = (dat >> 7);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 6);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 5);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 4);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 3);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 2);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 1);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 0);
  LCD_SCL_CLR;
  LCD_SCL_SET;

  LCD_CS_SET;

}

/**
  * @brief  spi write display data
  * @param  dat: write data
  * @retval none
  */
void spi4w_writedisplaydata(u16 dat)
{
  LCD_CS_CLR;
  LCD_RS_SET;

  LCD_SDA_SET;
  LCD_SCL_CLR;
  SPI_SDA = (dat >> 15);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 14);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 13);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 12);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 11);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 10);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 9);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 8);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 7);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 6);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 5);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 4);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 3);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 2);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 1);
  LCD_SCL_CLR;
  LCD_SCL_SET;
  SPI_SDA = (dat >> 0);
  LCD_SCL_CLR;
  LCD_SCL_SET;

  LCD_CS_SET;
}

/**
  * @brief  lcd read data
  * @param  none
  * @retval read data
  */
u16 LCD_SPI_RD_DATA(void)
{
  vu16 ram;

  ram = 0;

  LCD_CS_CLR;
  LCD_RS_SET;

  LCD_SCL_CLR;
  LCD_SDA_SET;

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 7);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 6);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 5);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 4);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 3);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 2);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 1);
  }

  LCD_SCL_CLR;
  LCD_SCL_SET;

  if(GPIOA->IPTDT & GPIO_IPTDT_IPTDT6)
  {
    ram |= (1 << 0);
  }

  LCD_CS_SET;

  return ram;
}
#endif //if(LCD_SPI1_EN)

/**
  * @brief  lcd write register
  * @param  regval: register value
  * @retval none
  */
void lcd_wr_reg(u16 regval)
{
  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_REG = regval >> 8;
    LCD->LCD_REG = regval & 0xff;
  }
  break;

  case 1:
  {
    LCD->LCD_REG = regval;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    lcd_spi1_sendindex((regval & 0xFF) >> 8);
    lcd_spi1_sendindex(regval & 0xFF);
#else
    spi4w_writeregindex((regval & 0xFF) >> 8);
    spi4w_writeregindex(regval & 0xFF);
#endif
    break;
  }

  default:
  {
    LCD->LCD_REG = regval;
  }
  break;
  }
}

/**
  * @brief  lcd write 8bit register
  * @param  regval: register value
  * @retval none
  */
void lcd_wr_8reg(u16 regval)
{
  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_REG = regval;
  }
  break;

  case 1:
  {
    LCD->LCD_REG = regval;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    lcd_spi1_sendindex(regval);
#else
    spi4w_writeregindex(regval);
#endif
    break;
  }

  default:
  {
    LCD->LCD_REG = regval;
  }
  break;
  }
}

/**
  * @brief  lcd read 8bit data
  * @param  none
  * @retval read data
  */
u8 lcd_rd_8data(void)
{
  u8 ram;
#if(LCD_SPI1_EN)
  ram = lcd_spi1_receivedata();
#else
  ram = LCD_SPI_RD_DATA();
#endif
  return ram;
}

/**
  * @brief  lcd write data
  * @param  data: data value
  * @retval none
  */
void lcd_wr_data(u16 data)
{
  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_RAM = data >> 8;
    LCD->LCD_RAM = data & 0xff;
  }
  break;

  case 1:
  {
    LCD->LCD_RAM = data;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    lcd_spi1_senddata(data);
    //SPI4W_WriteData1(data);
#else
    spi4w_writedata(data);
#endif
    break;
  }

  default:
  {
    LCD->LCD_RAM = data;
  }
  break;
  }

}

/**
  * @brief  lcd write 8bit data
  * @param  data: data value
  * @retval none
  */
void lcd_wr_8data(u16 data)
{
  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_RAM = data;
  }
  break;

  case 1:
  {
    LCD->LCD_RAM = data;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    lcd_spi1_senddata(data);
    //SPI4W_WriteData1(data);
#else
    spi4w_writedata(data);
#endif
    break;
  }

  default:
  {
    LCD->LCD_RAM = data;
  }
  break;
  }

}

/**
  * @brief  lcd write register data
  * @param  data1: register
  * @param  data2: data
  * @retval none
  */
void lcd_wr_reg_data(u16 data1, u16 data2)
{
  lcd_wr_8reg(data1);
  lcd_wr_8data(data2);
}

/**
  * @brief  lcd write register
  * @param  lcd_reg: register
  * @param  lcd_regvalue: register value
  * @retval none
  */
void lcd_writereg(u16 LCD_Reg, u16 LCD_RegValue)
{

  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_REG = LCD_Reg >> 8;
    LCD->LCD_REG = LCD_Reg & 0xff;
    LCD->LCD_RAM = LCD_RegValue >> 8;
    LCD->LCD_RAM = LCD_RegValue & 0xff;
  }
  break;

  case 1:
  {
    LCD->LCD_REG = LCD_Reg;
    LCD->LCD_RAM = LCD_RegValue;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    lcd_spi1_sendindex(LCD_Reg >> 8);
    lcd_spi1_sendindex(LCD_Reg & 0xff);
    SPI4W_WriteDisplayData1(LCD_RegValue >> 8);
    SPI4W_WriteDisplayData1(LCD_RegValue & 0xff);
#else
    spi4w_writeregindex(LCD_Reg >> 8);
    spi4w_writeregindex(LCD_Reg & 0xff);
    spi4w_writedisplaydata(LCD_RegValue >> 8);
    spi4w_writedisplaydata(LCD_RegValue & 0xff);
#endif
    break;
  }

  case 4:
  {
    lcd_wr_reg(LCD_Reg);
    lcd_wr_data(LCD_RegValue);
  }
  break;

  default:
  {
    LCD->LCD_REG = LCD_Reg;
    LCD->LCD_RAM = LCD_RegValue;
  }
  break;
  }
}

/**
  * @brief  lcd read data
  * @param  none
  * @retval read data value
  */
u16 lcd_rd_data(void)
{
  vu16 ram;
  ram = LCD->LCD_RAM;
  return ram;
}

/**
  * @brief  lcd read register data
  * @param  lcd_reg: register
  * @retval read register value
  */
u16 lcd_readreg(u16 LCD_Reg)
{
  switch(MCU)
  {
  case 0:
  {
    lcd_wr_reg(LCD_Reg);
    delay_us(5);
    return lcd_rd_data();
  }

  case 1:
  {
    u16 info;
    lcd_wr_reg(LCD_Reg);
    delay_us(5);
    info = lcd_rd_data();
    info <<= 8;
    info |= lcd_rd_data();
    return info;
  }

  case 2:
  {	 			}			break;

  case 3:
  {	 			}			break;

  default:
    break;
  }

  return 0;
}

/**
  * @brief  set lcd scan direction
  * @param  dir: direction
  *         0: 0, 1: 90, 2: 180, 3: 270
  * @retval none
  */
void lcd_scan_dir(u8 dir)
{
  switch(dir)
  {
  case 0:
  {
    lcddev.width = 240;
    lcddev.height = 320;
    lcddev.setxcmd = 0X2A;
    lcddev.setycmd = 0X2B;
    lcddev.wramcmd = 0X2C;

    lcd_wr_8reg(0x36);
    lcd_wr_8data(0x08);
  }
  break;

  case 1:
  {
    lcddev.width = 320;
    lcddev.height = 240;
    lcddev.setxcmd = 0X2A;
    lcddev.setycmd = 0X2B;
    lcddev.wramcmd = 0X2C;

    lcd_wr_8reg(0x36);
    lcd_wr_8data(0xA8);
  }
  break;

  case 2:
  {
    lcddev.width = 240;
    lcddev.height = 320;
    lcddev.setxcmd = 0X2A;
    lcddev.setycmd = 0X2B;
    lcddev.wramcmd = 0X2C;

    lcd_wr_8reg(0x36);
    lcd_wr_8data(0xC8);
  }
  break;

  case 3:
  {
    lcddev.width = 320;
    lcddev.height = 240;
    lcddev.setxcmd = 0X2A;
    lcddev.setycmd = 0X2B;
    lcddev.wramcmd = 0X2C;

    lcd_wr_8reg(0x36);
    lcd_wr_8data(0x68);
  }
  break;

  }
}

/**
  * @brief  write lcd block
  * @param  xstart: x direction start
  * @param  xend: x direction end
  * @param  ystart: y direction start
  * @param  yend: y direction end
  * @retval none
  */
void blockwrite(unsigned int Xstart, unsigned int Xend, unsigned int Ystart, unsigned int Yend)
{
  lcd_wr_8reg(lcddev.setxcmd);
  lcd_wr_8data(Xstart >> 8);
  lcd_wr_8data(Xstart & 0xff);
  lcd_wr_8data(Xend >> 8);
  lcd_wr_8data(Xend & 0xff);

  lcd_wr_8reg(lcddev.setycmd);
  lcd_wr_8data(Ystart >> 8);
  lcd_wr_8data(Ystart & 0xff);
  lcd_wr_8data(Yend >> 8);
  lcd_wr_8data(Yend & 0xff);

  lcd_wr_8reg(0x2C);
}

/**
  * @brief  set lcd window
  * @param  sx: x direction start
  * @param  sy: y direction start
  * @param  width: window width
  * @param  height: window height
  * @retval none
  */
void lcd_set_window(u16 sx, u16 sy, u16 width, u16 height)
{
  u8 hsareg, heareg, vsareg, veareg;
  u16 hsaval, heaval, vsaval, veaval;
  u16 twidth, theight;
  twidth = sx + width - 1;
  theight = sy + height - 1;

  if(lcddev.id == 0X9341 || lcddev.id == 0X5310 || lcddev.id == 0X6804 || (lcddev.dir == 1 && lcddev.id == 0X1963))
  {
    lcd_wr_reg(lcddev.setxcmd);
    lcd_wr_data(sx >> 8);
    lcd_wr_data(sx & 0XFF);
    lcd_wr_data(twidth >> 8);
    lcd_wr_data(twidth & 0XFF);
    lcd_wr_reg(lcddev.setycmd);
    lcd_wr_data(sy >> 8);
    lcd_wr_data(sy & 0XFF);
    lcd_wr_data(theight >> 8);
    lcd_wr_data(theight & 0XFF);
  }
  else if(lcddev.id == 0X1963)
  {
    sx = lcddev.width - width - sx;
    height = sy + height - 1;
    lcd_wr_reg(lcddev.setxcmd);
    lcd_wr_data(sx >> 8);
    lcd_wr_data(sx & 0XFF);
    lcd_wr_data((sx + width - 1) >> 8);
    lcd_wr_data((sx + width - 1) & 0XFF);
    lcd_wr_reg(lcddev.setycmd);
    lcd_wr_data(sy >> 8);
    lcd_wr_data(sy & 0XFF);
    lcd_wr_data(height >> 8);
    lcd_wr_data(height & 0XFF);
  }
  else if(lcddev.id == 0X5510)
  {
    lcd_wr_reg(lcddev.setxcmd);
    lcd_wr_data(sx >> 8);
    lcd_wr_reg(lcddev.setxcmd + 1);
    lcd_wr_data(sx & 0XFF);
    lcd_wr_reg(lcddev.setxcmd + 2);
    lcd_wr_data(twidth >> 8);
    lcd_wr_reg(lcddev.setxcmd + 3);
    lcd_wr_data(twidth & 0XFF);
    lcd_wr_reg(lcddev.setycmd);
    lcd_wr_data(sy >> 8);
    lcd_wr_reg(lcddev.setycmd + 1);
    lcd_wr_data(sy & 0XFF);
    lcd_wr_reg(lcddev.setycmd + 2);
    lcd_wr_data(theight >> 8);
    lcd_wr_reg(lcddev.setycmd + 3);
    lcd_wr_data(theight & 0XFF);
  }
  else
  {
    if(lcddev.dir == 1)
    {
      hsaval = sy;
      heaval = theight;
      vsaval = lcddev.width - twidth - 1;
      veaval = lcddev.width - sx - 1;
    }
    else
    {
      hsaval = sx;
      heaval = twidth;
      vsaval = sy;
      veaval = theight;
    }

    hsareg = 0X50;
    heareg = 0X51;
    vsareg = 0X52;
    veareg = 0X53;

    lcd_writereg(hsareg, hsaval);
    lcd_writereg(heareg, heaval);
    lcd_writereg(vsareg, vsaval);
    lcd_writereg(veareg, veaval);
    lcd_setcursor(sx, sy);
  }
}

/**
  * @brief  set lcd cursor position
  * @param  sx: x position
  * @param  sy: y position
  * @retval none
  */
void lcd_setcursor(u16 Xpos, u16 Ypos)
{
  lcd_wr_8reg(lcddev.setxcmd);
  lcd_wr_8data(Xpos >> 8);
  lcd_wr_8data(Xpos & 0xff);

  lcd_wr_8reg(lcddev.setycmd);
  lcd_wr_8data(Ypos >> 8);
  lcd_wr_8data(Ypos & 0xff);

  lcd_wr_8reg(0x2C);
}

/**
  * @brief  read lcd id
  * @param  none
  * @retval none
  */
void LCDILI9341_ReadID(void)
{
  lcd_wr_8reg(0XD9);
  lcd_wr_8data(0x11);
  lcd_wr_8reg(0XD3);
  delay_ms(1);
  lcddev.id = lcd_rd_8data();

  lcd_wr_8reg(0XD9);
  lcd_wr_8data(0x12);
  lcd_wr_reg(0XD3);
  delay_ms(1);
  lcddev.id = lcd_rd_8data();


  lcd_wr_8reg(0XD9);
  lcd_wr_8data(0x13);
  lcd_wr_reg(0XD3);
  delay_ms(1);
  lcddev.id <<= 8;
  lcddev.id |= lcd_rd_8data();
}

/**
  * @brief  prepare to write lcd ram
  * @param  none
  * @retval none
  */
void lcd_writeram_prepare(void)
{
  lcd_wr_8reg(lcddev.wramcmd);
}

/**
  * @brief  lcd write ram
  * @param  rgb_code: color value code
  * @retval none
  */
void lcd_writeram(u16 RGB_Code)
{
  switch(MCU)
  {
  case 0:
  {
    LCD->LCD_RAM = RGB_Code >> 8;
    LCD->LCD_RAM = RGB_Code & 0xff;
  }
  break;

  case 1:
  {
    LCD->LCD_RAM = RGB_Code;
  }
  break;

  case 2:
  {
#if(LCD_SPI1_EN)
    SPI4W_WriteDisplayData1(RGB_Code);
#else
    spi4w_writedisplaydata(RGB_Code);
#endif
    break;
  }

  default:
  {
    LCD->LCD_RAM = RGB_Code;
  }
  break;
  }
}

/**
  * @brief  set lcd qvga mode
  * @param  none
  * @retval none
  */
void LCD_QVGA(void)
{
  lcddev.width = 320;
  lcddev.height = 240;
  lcddev.setxcmd = 0X2A;
  lcddev.setycmd = 0X2B;
  lcddev.wramcmd = 0X2C;

}

/**
  * @brief  set lcd register parameters
  * @param  none
  * @retval none
  */
void LCD_Set(void)
{
  LCD_QVGA()	;	//240RGB320
}

#define LCD_ILI9341_CMD  lcd_wr_8reg
#define LCD_ILI9341_DATA  lcd_wr_8data
#define delayms  delay_ms

/**
  * @brief  initialize lcd ili9341
  * @param  none
  * @retval none
  */
void ili9341_ivo24_initial(void)
{
  LCD_ILI9341_CMD(0xCF);
  LCD_ILI9341_DATA(0x00);
  LCD_ILI9341_DATA(0xC1);
  LCD_ILI9341_DATA(0X30);

  LCD_ILI9341_CMD(0xED);
  LCD_ILI9341_DATA(0x64);
  LCD_ILI9341_DATA(0x03);
  LCD_ILI9341_DATA(0X12);
  LCD_ILI9341_DATA(0X81);

  LCD_ILI9341_CMD(0xE8);
  LCD_ILI9341_DATA(0x85);
  LCD_ILI9341_DATA(0x10);
  LCD_ILI9341_DATA(0x78);

  LCD_ILI9341_CMD(0xCB);
  LCD_ILI9341_DATA(0x39);
  LCD_ILI9341_DATA(0x2C);
  LCD_ILI9341_DATA(0x00);
  LCD_ILI9341_DATA(0x34);
  LCD_ILI9341_DATA(0x02);

  LCD_ILI9341_CMD(0xF7);
  LCD_ILI9341_DATA(0x20);

  LCD_ILI9341_CMD(0xEA);
  LCD_ILI9341_DATA(0x00);
  LCD_ILI9341_DATA(0x00);

  LCD_ILI9341_CMD(0xC0);    //power control
  LCD_ILI9341_DATA(0x21);   //vrh[5:0]

  LCD_ILI9341_CMD(0xC1);    //power control
  LCD_ILI9341_DATA(0x12);   //sap[2:0];bt[3:0]

  LCD_ILI9341_CMD(0xC5);    //vcm control
  LCD_ILI9341_DATA(0x32);
  LCD_ILI9341_DATA(0x3C);

  LCD_ILI9341_CMD(0xC7);    //vcm control2
  LCD_ILI9341_DATA(0XA7);   //b4

  LCD_ILI9341_CMD(0x36);    // memory access control
  LCD_ILI9341_DATA(0x08);

  LCD_ILI9341_CMD(0x3A);
  LCD_ILI9341_DATA(0x55);

  LCD_ILI9341_CMD(0xB1);
  LCD_ILI9341_DATA(0x00);
  LCD_ILI9341_DATA(0x17);

  LCD_ILI9341_CMD(0xB6);    // display function control
  LCD_ILI9341_DATA(0x0A);
  LCD_ILI9341_DATA(0xA2);

  LCD_ILI9341_CMD(0xF6);
  LCD_ILI9341_DATA(0x01);
  LCD_ILI9341_DATA(0x30);

  LCD_ILI9341_CMD(0xF2);    // 3gamma function disable
  LCD_ILI9341_DATA(0x00);

  LCD_ILI9341_CMD(0x26);    //gamma curve selected
  LCD_ILI9341_DATA(0x01);

  LCD_ILI9341_CMD(0xE0);    //set gamma
  LCD_ILI9341_DATA(0x0F);
  LCD_ILI9341_DATA(0x20);
  LCD_ILI9341_DATA(0x1E);
  LCD_ILI9341_DATA(0x07);
  LCD_ILI9341_DATA(0x0A);
  LCD_ILI9341_DATA(0x03);
  LCD_ILI9341_DATA(0x52);
  LCD_ILI9341_DATA(0X63);
  LCD_ILI9341_DATA(0x44);
  LCD_ILI9341_DATA(0x08);
  LCD_ILI9341_DATA(0x17);
  LCD_ILI9341_DATA(0x09);
  LCD_ILI9341_DATA(0x19);
  LCD_ILI9341_DATA(0x13);
  LCD_ILI9341_DATA(0x00);

  LCD_ILI9341_CMD(0XE1);    //set gamma
  LCD_ILI9341_DATA(0x00);
  LCD_ILI9341_DATA(0x16);
  LCD_ILI9341_DATA(0x19);
  LCD_ILI9341_DATA(0x02);
  LCD_ILI9341_DATA(0x0F);
  LCD_ILI9341_DATA(0x03);
  LCD_ILI9341_DATA(0x2F);
  LCD_ILI9341_DATA(0x13);
  LCD_ILI9341_DATA(0x40);
  LCD_ILI9341_DATA(0x01);
  LCD_ILI9341_DATA(0x08);
  LCD_ILI9341_DATA(0x07);
  LCD_ILI9341_DATA(0x2E);
  LCD_ILI9341_DATA(0x3C);
  LCD_ILI9341_DATA(0x0F);

  LCD_ILI9341_CMD(0x11);    //exit sleep
  delay_ms(520);
  LCD_ILI9341_CMD(0x29);    //display on
}

/**
  * @brief  initialize lcd
  * @param  none
  * @retval none
  */
void lcd_init(void)
{
  lcd_ctrllinesconfig();
  lcd_reset();

  delay_ms(5); 					// delay 5 ms

  LCDILI9341_ReadID();
  LCD_Set();

  ili9341_ivo24_initial();
}

/**
  * @brief  convert bgr to rgb
  * @param  c: convert bgr value
  * @retval rgb value
  */
u16 lcd_bgr2rgb(u16 c)
{
  u16  r, g, b, rgb;
  b = (c >> 0) & 0x1f;
  g = (c >> 5) & 0x3f;
  r = (c >> 11) & 0x1f;
  rgb = (b << 11) + (g << 5) + (r << 0);
  return(rgb);
}

/**
  * @brief  operation delay
  * @param  i: delay value
  * @retval none
  */
void opt_delay(u8 i)
{
  while(i--);
}

/**
  * @brief  lcd read point color value
  * @param  x: x coordinates
  * @param  y: y coordinates
  * @retval point color value
  */
u16 lcd_readpoint(u16 x, u16 y)
{
  u16 r = 0, g = 0, b = 0;

  if(x >= lcddev.width || y >= lcddev.height)
  {
    return 0;
  }

  lcd_setcursor(x, y);

  if(lcddev.id == 0X9341 || lcddev.id == 0X6804 || lcddev.id == 0X5310 || lcddev.id == 0X1963)
  {
    lcd_wr_reg(0X2E);
  }
  else if(lcddev.id == 0X5510)
  {
    lcd_wr_reg(0X2E00);
  }
  else
  {
    lcd_wr_reg(0X22);
  }

  if(lcddev.id == 0X9320)
  {
    opt_delay(2);
  }

  r = lcd_rd_data();        //dummy Read

  if(lcddev.id == 0X1963)
  {
    return r;
  }

  opt_delay(2);
  r = lcd_rd_data();

  if(lcddev.id == 0X9341 || lcddev.id == 0X5310 || lcddev.id == 0X5510)
  {
    opt_delay(2);
    b = lcd_rd_data();
    g = r & 0XFF;
    g <<= 8;
  }

  if(lcddev.id == 0X9325 || lcddev.id == 0X4535 || lcddev.id == 0X4531 || lcddev.id == 0XB505 || lcddev.id == 0XC505)
  {
    return r;
  }
  else if(lcddev.id == 0X9341 || lcddev.id == 0X5310 || lcddev.id == 0X5510)
  {
    return (((r >> 11) << 11) | ((g >> 10) << 5) | (b >> 11));
  }
  else
  {
    return lcd_bgr2rgb(r);
  }
}

/**
  * @brief  clear lcd
  * @param  color: clear lcd with color
  * @retval none
  */
void lcd_clear(u16 color)
{
  u32 index = 0;
  u32 totalpoint = lcddev.width;
  totalpoint *= lcddev.height;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(index = 0; index < totalpoint; index++)
  {
    lcd_writeram(color);
  }
}

/**
  * @brief  clear part of lcd
  * @param  color: clear lcd with color
  * @retval none
  */
void lcd_clear_part(u16 color)
{
  u32 index = 0;
  u32 totalpoint = lcddev.width;
  totalpoint *= lcddev.height - 272;

  blockwrite(0, lcddev.width, 32, 80 - 1);

  for(index = 0; index < totalpoint; index++)
  {
    lcd_writeram(color);
    lcd_writeram(color);
  }
}

/**
  * @brief  display lcd
  * @param  color: display lcd with color
  * @retval none
  */
void lcd_display(u16 color)
{
  u16 i, j;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(i = 0; i < lcddev.height; i++)
  {
    for(j = 0; j < lcddev.width; j++)
    {
      lcd_writeram(color);
    }
  }
}

/**
  * @brief  lcd frame display
  * @param  rgb_data: middle fill color
  * @param  f_data: frame color
  * @retval none
  */
void lcd_frame(u16 RGB_Data, u16 f_data)
{
  u16 i, j;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(i = 0; i < 1; i++)
  {
    for(j = 0; j < lcddev.width; j++)
    {
      lcd_writeram(f_data);
    }
  }

  for (i = 0; i < (lcddev.height - 2); i++)
  {
    for (j = 0; j < 1; j++)
    {
      lcd_writeram(f_data);
    }

    for (j = 0; j < (lcddev.width - 2); j++)
    {
      lcd_writeram(RGB_Data);
    }

    for (j = 0; j < 1; j++)
    {
      lcd_writeram(f_data);
    }
  }

  for(i = 0; i < 1; i++)
  {
    for(j = 0; j < lcddev.width; j++)
    {
      lcd_writeram(f_data);
    }
  }
}

/**
  * @brief  lcd show gray
  * @param  none
  * @retval none
  */
void show_gray(void)
{
  unsigned int i, j, k, c16h;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(j = 0; j < (lcddev.height); j++)
  {
    c16h = 0x0000;

    for(k = 0; k < 16; k++)
    {
      for(i = 0; i < (lcddev.width / 16); i++)
      {
        lcd_writeram(c16h);
      }

      c16h = c16h + 0x1082;
    }
  }
}

/**
  * @brief  check if crosstalk
  * @param  none
  * @retval none
  */
void check_crosstalk(void)
{
  unsigned short m, n, i, j;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(m = 0; m < lcddev.height / 4; m++)
  {
    for(n = 0; n < lcddev.width; n++)
    {
      lcd_writeram(0x8410);
    }
  }

  for(m = 0; m < lcddev.height / 2; m++)
  {
    for(n = 0; n < lcddev.width / 4; n++)
    {
      lcd_writeram(0x8410);
    }

    for(i = 0; i < lcddev.width / 2; i++)
    {
      lcd_writeram(0x0000);
    }

    for(j = 0; j < lcddev.width / 4; j++)
    {
      lcd_writeram(0x8410);
    }
  }

  for(m = 0; m < lcddev.height / 4; m++)
  {
    for(n = 0; n < lcddev.width; n++)
    {
      lcd_writeram(0x8410);
    }
  }
}

/**
  * @brief  check if Flicker
  * @param  none
  * @retval none
  */
void check_flicker(void)
{
  unsigned short i, j;

  blockwrite(0, lcddev.width, 0, lcddev.height);

  for(i = 0; i < lcddev.height; i++)
  {
    for(j = 0; j < lcddev.width; j++)
    {
      lcd_writeram(0x8410);
    }

    i++;

    for(j = 0; j < lcddev.width; j++)
    {
      lcd_writeram(0x0000);
    }
  }
}

/**
  * @brief  fill lcd in designated area
  * @param  sx: x direction start
  * @param  sy: y direction start
  * @param  ex: x direction end
  * @param  ey: y direction end
  * @param  color: fill color
  * @retval none
  */
void lcd_fill(u16 sx, u16 sy, u16 ex, u16 ey, u16 color)
{
  u16 i, j;
  u16 xlen = 0;

  xlen = ex - sx + 1;

  for(i = sy; i <= ey; i++)
  {
    lcd_setcursor(sx, i);
    lcd_writeram_prepare();

    for(j = 0; j < xlen; j++)
    {
      lcd_writeram(color);
    }
  }
}

/**
  * @brief  fill lcd in designated area with different color
  * @param  sx: x direction start
  * @param  sy: y direction start
  * @param  ex: x direction end
  * @param  ey: y direction end
  * @param  color: fill color point
  * @retval none
  */
void lcd_color_fill(u16 sx, u16 sy, u16 ex, u16 ey, u16 *color)
{
  u16 height, width;
  u16 i, j;
  width = ex - sx + 1;
  height = ey - sy + 1;

  for(i = 0; i < height; i++)
  {
    lcd_setcursor(sx, sy + i);
    lcd_writeram_prepare();

    for(j = 0; j < width; j++)
    {
      lcd_writeram(color[i * width + j]);
    }
  }
}

/**
  * @brief  lcd draw a point
  * @param  x: x coordinate value
  * @param  y: y coordinate value
  * @retval none
  */
void lcd_drawpoint(u16 x, u16 y)
{
  lcd_setcursor(x, y);

  lcd_writeram(POINT_COLOR);
}

/**
  * @brief  lcd draw a big point
  * @param  x: x coordinate value
  * @param  y: y coordinate value
  * @retval none
  */
void lcd_drawpoint_big(u16 x, u16 y)
{
  lcd_fill(x - 1, y - 1, x + 1, y + 1, POINT_COLOR);
}

/**
  * @brief  lcd draw a point fast
  * @param  x: x coordinate value
  * @param  y: y coordinate value
  * @param  color: draw color
  * @retval none
  */
void lcd_fast_drawpoint(u16 x, u16 y, u16 color)
{
  lcd_setcursor(x, y);

  lcd_writeram(color);
}

/**
  * @brief  lcd draw a line
  * @param  x1: start x coordinate value
  * @param  y1: start y coordinate value
  * @param  x2: end x coordinate value
  * @param  y2: end y coordinate value
  * @retval none
  */
void lcd_drawline(u16 x1, u16 y1, u16 x2, u16 y2)
{
  u16 t;
  int xerr = 0, yerr = 0, delta_x, delta_y, distance;
  int incx, incy, uRow, uCol;
  delta_x = x2 - x1;
  delta_y = y2 - y1;
  uRow = x1;
  uCol = y1;

  if(delta_x > 0)
  {
    incx = 1;
  }
  else if(delta_x == 0)
  {
    incx = 0;  //perpendicular line
  }
  else
  {
    incx = -1;
    delta_x = -delta_x;
  }

  if(delta_y > 0)
  {
    incy = 1;
  }
  else if(delta_y == 0)
  {
    incy = 0;  //horizontal line
  }
  else
  {
    incy = -1;
    delta_y = -delta_y;
  }

  if( delta_x > delta_y)
  {
    distance = delta_x;
  }
  else
  {
    distance = delta_y;
  }

  for(t = 0; t <= distance + 1; t++ )
  {
    lcd_drawpoint(uRow, uCol);
    xerr += delta_x ;
    yerr += delta_y ;

    if(xerr > distance)
    {
      xerr -= distance;
      uRow += incx;
    }

    if(yerr > distance)
    {
      yerr -= distance;
      uCol += incy;
    }
  }
}

/**
  * @brief  lcd draw a rectangle
  * @param  x1: start x coordinate value
  * @param  y1: start y coordinate value
  * @param  x2: end x coordinate value
  * @param  y2: end y coordinate value
  * @retval none
  */
void lcd_drawrectangle(u16 x1, u16 y1, u16 x2, u16 y2)
{
  lcd_drawline(x1, y1, x2, y1);
  lcd_drawline(x1, y1, x1, y2);
  lcd_drawline(x1, y2, x2, y2);
  lcd_drawline(x2, y1, x2, y2);
}

/**
  * @brief  lcd draw a circle
  * @param  x0: center of circle x coordinate value
  * @param  y0: center of circle y coordinate value
  * @param  r: radius value
  * @retval none
  */
void lcd_draw_circle(u16 x0, u16 y0, u8 r)
{
  int a, b;
  int di;
  a = 0;
  b = r;
  di = 3 - (r << 1);

  while(a <= b)
  {
    lcd_drawpoint(x0 + a, y0 - b);        //5
    lcd_drawpoint(x0 + b, y0 - a);        //0
    lcd_drawpoint(x0 + b, y0 + a);        //4
    lcd_drawpoint(x0 + a, y0 + b);        //6
    lcd_drawpoint(x0 - a, y0 + b);        //1
    lcd_drawpoint(x0 - b, y0 + a);
    lcd_drawpoint(x0 - a, y0 - b);        //2
    lcd_drawpoint(x0 - b, y0 - a);        //7
    a++;

    /* draw a circle with bresenham algorithm */
    if(di < 0)
    {
      di += 4 * a + 6;
    }
    else
    {
      di += 10 + 4 * (a - b);
      b--;
    }
  }
}

/**
  * @brief  lcd show a character
  * @param  x: start x coordinate value
  * @param  y: start y coordinate value
  * @param  num: character number
  * @param  size: character size
  * @param  mode: superposition mode
  * @retval none
  */
void lcd_showchar(u16 x, u16 y, u8 num, u8 size, u8 mode)
{
  u8 temp, t1, t;
  u16 y0 = y;
  u8 csize = (size / 8 + ((size % 8) ? 1 : 0)) * (size / 2);
  num = num - ' ';

  for(t = 0; t < csize; t++)
  {
    if(size == 12)
    {
      temp = asc2_1206[num][t];  //1206 typeface
    }
    else if(size == 16)
    {
      temp = asc2_1608[num][t];  //1608 typeface
    }
    else if(size == 24)
    {
      temp = asc2_2412[num][t];  //2412 typeface
    }
    else
    {
      return;
    }

    for(t1 = 0; t1 < 8; t1++)
    {
      if(temp & 0x80)
      {
        lcd_fast_drawpoint(x, y, POINT_COLOR);
      }
      else if(mode == 0)
      {
        lcd_fast_drawpoint(x, y, BACK_COLOR);
      }

      temp <<= 1;
      y++;

      if(y >= lcddev.height)
      {
        return;
      }

      if((y - y0) == size)
      {
        y = y0;
        x++;

        if(x >= lcddev.width)
        {
          return;
        }

        break;
      }
    }
  }
}

/**
  * @brief  lcd power
  * @param  m: the base of the power
  * @param  n: the exponential of the power
  * @retval result
  */
u32 lcd_pow(u8 m, u8 n)
{
  u32 result = 1;

  while(n--)
  {
    result *= m;
  }

  return result;
}

/**
  * @brief  lcd show a number
  * @param  x: start x coordinate value
  * @param  y: start y coordinate value
  * @param  num: number value
  * @param  len: number length
  * @param  size: number size
  * @retval none
  */
void lcd_shownum(u16 x, u16 y, u32 num, u8 len, u8 size)
{
  u8 t, temp;
  u8 enshow = 0;

  for(t = 0; t < len; t++)
  {
    temp = (num / lcd_pow(10, len - t - 1)) % 10;

    if(enshow == 0 && t < (len - 1))
    {
      if(temp == 0)
      {
        lcd_showchar(x + (size / 2)*t, y, ' ', size, 0);
        continue;
      }
      else
      {
        enshow = 1;
      }

    }

    lcd_showchar(x + (size / 2)*t, y, temp + '0', size, 0);
  }
}

/**
  * @brief  lcd show numbers
  * @param  x: start x coordinate value
  * @param  y: start y coordinate value
  * @param  num: number value
  * @param  len: number length
  * @param  size: number size
  * @param  mode: superposition mode
  * @retval none
  */
void lcd_showxnum(u16 x, u16 y, u32 num, u8 len, u8 size, u8 mode)
{
  u8 t, temp;
  u8 enshow = 0;

  for(t = 0; t < len; t++)
  {
    temp = (num / lcd_pow(10, len - t - 1)) % 10;

    if(enshow == 0 && t < (len - 1))
    {
      if(temp == 0)
      {
        if(mode & 0X80)
        {
          lcd_showchar(x + (size / 2)*t, y, '0', size, mode & 0X01);
        }
        else
        {
          lcd_showchar(x + (size / 2)*t, y, ' ', size, mode & 0X01);
        }

        continue;
      }
      else
      {
        enshow = 1;
      }

    }

    lcd_showchar(x + (size / 2)*t, y, temp + '0', size, mode & 0X01);
  }
}

/**
  * @brief  lcd show string
  * @param  x: start x coordinate value
  * @param  y: start y coordinate value
  * @param  width: string width
  * @param  height: string height
  * @param  size: string size
  * @param  p: string point
  * @retval none
  */
void lcd_showstring(u16 x, u16 y, u16 width, u16 height, u8 size, u8 *p)
{
  u8 x0 = x;
  width += x;
  height += y;

  while((*p <= '~') && (*p >= ' '))
  {
    if(x >= width)
    {
      x = x0;
      y += size;
    }

    if(y >= height)
    {
      break;
    }

    lcd_showchar(x, y, *p, size, 0);
    x += size / 2;
    p++;
  }
}
